@@ -1,7 +1,7 @@
 import json
 from abc import ABC, abstractmethod
 from base64 import b64decode, b64encode
-from typing import Generic, List
+from typing import Generic, List, Optional, Sequence
 from uuid import UUID
 
 from eventsourcing.application import NotificationLog, Section, TApplication
@@ -22,7 +22,9 @@ class NotificationLogInterface(ABC):
         """
 
     @abstractmethod
-    def get_notifications(self, start: int, limit: int) -> str:
+    def get_notifications(
+        self, start: int, limit: int, topics: Sequence[str] = ()
+    ) -> str:
         """
         Returns a serialised list of :class:`~eventsourcing.persistence.Notification`
         objects from a notification log.
@@ -45,7 +47,7 @@ class NotificationLogJSONService(NotificationLogInterface, Generic[TApplication]
         Returns JSON serialised :class:`~eventsourcing.application.Section`
         from a notification log.
         """
-        section = self.app.log[section_id]
+        section = self.app.notification_log[section_id]
         return json.dumps(
             {
                 "id": section.id,
@@ -63,8 +65,12 @@ class NotificationLogJSONService(NotificationLogInterface, Generic[TApplication]
             }
         )
 
-    def get_notifications(self, start: int, limit: int) -> str:
-        notifications = self.app.log.select(start, limit)
+    def get_notifications(
+        self, start: int, limit: int, topics: Sequence[str] = ()
+    ) -> str:
+        notifications = self.app.notification_log.select(
+            start=start, limit=limit, topics=topics
+        )
         return json.dumps(
             [
                 {
@@ -113,7 +119,13 @@ class NotificationLogJSONClient(NotificationLog):
             ],
         )
 
-    def select(self, start: int, limit: int) -> List[Notification]:
+    def select(
+        self,
+        start: int,
+        limit: int,
+        stop: Optional[int] = None,
+        topics: Sequence[str] = (),
+    ) -> List[Notification]:
         """
         Returns a selection
         :class:`~eventsourcing.persistence.Notification` objects
@@ -127,5 +139,9 @@ class NotificationLogJSONClient(NotificationLog):
                 topic=item["topic"],
                 state=b64decode(item["state"].encode("utf8")),
             )
-            for item in json.loads(self.interface.get_notifications(start, limit))
+            for item in json.loads(
+                self.interface.get_notifications(
+                    start=start, limit=limit, topics=topics
+                )
+            )
         ]
