@@ -1,14 +1,16 @@
 import sqlite3
-import threading
+from contextlib import contextmanager
 from sqlite3 import Connection, Cursor
-from threading import Lock
 from types import TracebackType
-from typing import Any, Dict, List, Mapping, Optional, Type
+from typing import Any, Iterator, List, Optional, Sequence, Type, Union, cast
 from uuid import UUID
 
 from eventsourcing.persistence import (
     AggregateRecorder,
     ApplicationRecorder,
+    Connection as BaseConnection,
+    ConnectionPool,
+    Cursor as BaseCursor,
     DatabaseError,
     DataError,
     InfrastructureFactory,
@@ -24,27 +26,76 @@ from eventsourcing.persistence import (
     StoredEvent,
     Tracking,
 )
-from eventsourcing.utils import strtobool
+from eventsourcing.utils import Environment, strtobool
 
 SQLITE3_DEFAULT_LOCK_TIMEOUT = 5
 
 
-class Transaction:
-    def __init__(
-        self, connection: Connection, commit: bool = False, lock: Optional[Lock] = None
-    ):
+class SQLiteCursor(BaseCursor):
+    def __init__(self, sqlite_cursor: Cursor):
+        self.sqlite_cursor = sqlite_cursor
+
+    def __enter__(self, *args: Any, **kwargs: Any) -> Cursor:
+        return self.sqlite_cursor
+
+    def __exit__(self, *args: Any, **kwargs: Any) -> None:
+        self.sqlite_cursor.close()
+
+    def execute(self, *args: Any, **kwargs: Any) -> None:
+        self.sqlite_cursor.execute(*args, **kwargs)
+
+    def executemany(self, *args: Any, **kwargs: Any) -> None:
+        self.sqlite_cursor.executemany(*args, **kwargs)
+
+    def fetchall(self) -> Any:
+        return self.sqlite_cursor.fetchall()
+
+    def fetchone(self) -> Any:
+        return self.sqlite_cursor.fetchone()
+
+    @property
+    def lastrowid(self) -> Any:
+        return self.sqlite_cursor.lastrowid
+
+
+class SQLiteConnection(BaseConnection[SQLiteCursor]):
+    def __init__(self, sqlite_conn: Connection, max_age: Optional[float]):
+        super().__init__(max_age=max_age)
+        self._sqlite_conn = sqlite_conn
+
+    @contextmanager
+    def transaction(self, commit: bool) -> Iterator[SQLiteCursor]:
+        # Context managed transaction.
+        with SQLiteTransaction(self, commit) as curs:
+            # Context managed cursor.
+            with curs:
+                yield curs
+
+    def cursor(self) -> SQLiteCursor:
+        return SQLiteCursor(self._sqlite_conn.cursor())
+
+    def rollback(self) -> None:
+        self._sqlite_conn.rollback()
+
+    def commit(self) -> None:
+        self._sqlite_conn.commit()
+
+    def _close(self) -> None:
+        self._sqlite_conn.close()
+        super()._close()
+
+
+class SQLiteTransaction:
+    def __init__(self, connection: SQLiteConnection, commit: bool = False):
         self.connection = connection
         self.commit = commit
-        self.lock = lock
 
-    def __enter__(self) -> Cursor:
-        if self.lock:
-            self.lock.acquire()
+    def __enter__(self) -> SQLiteCursor:
         # We must issue a "BEGIN" explicitly
         # when running in auto-commit mode.
-        self.connection.execute("BEGIN")
-        self.cursor = self.connection.cursor()
-        return self.cursor
+        cursor = self.connection.cursor()
+        cursor.execute("BEGIN")
+        return cursor
 
     def __exit__(
         self,
@@ -53,7 +104,6 @@ class Transaction:
         exc_tb: TracebackType,
     ) -> None:
         try:
-            self.cursor.close()
             if exc_val:
                 # Roll back all changes
                 # if an exception occurs.
@@ -64,58 +114,55 @@ class Transaction:
             else:
                 self.connection.commit()
         except sqlite3.InterfaceError as e:
-            raise InterfaceError(e)
+            raise InterfaceError(e) from e
         except sqlite3.DataError as e:
-            raise DataError(e)
+            raise DataError(e) from e
         except sqlite3.OperationalError as e:
-            raise OperationalError(e)
+            raise OperationalError(e) from e
         except sqlite3.IntegrityError as e:
-            raise IntegrityError(e)
+            raise IntegrityError(e) from e
         except sqlite3.InternalError as e:
-            raise InternalError(e)
+            raise InternalError(e) from e
         except sqlite3.ProgrammingError as e:
-            raise ProgrammingError(e)
+            raise ProgrammingError(e) from e
         except sqlite3.NotSupportedError as e:
-            raise NotSupportedError(e)
+            raise NotSupportedError(e) from e
         except sqlite3.DatabaseError as e:
-            raise DatabaseError(e)
+            raise DatabaseError(e) from e
         except sqlite3.Error as e:
-            raise PersistenceError(e)
-        finally:
-            if self.lock:
-                self.lock.release()
+            raise PersistenceError(e) from e
 
 
-class SQLiteDatastore:
-    def __init__(self, db_name: str, lock_timeout: Optional[int] = None):
+class SQLiteConnectionPool(ConnectionPool[SQLiteConnection]):
+    def __init__(
+        self,
+        db_name: str,
+        lock_timeout: Optional[int] = None,
+        pool_size: int = 5,
+        max_overflow: int = 10,
+        pool_timeout: float = 5.0,
+        max_age: Optional[float] = None,
+        pre_ping: bool = False,
+    ):
         self.db_name = db_name
+        self.lock_timeout = lock_timeout
         self.is_sqlite_memory_mode = self.detect_memory_mode(db_name)
-        self.lock: Optional[Lock] = None
-        if self.is_sqlite_memory_mode:
-            self.lock = Lock()
-
-        self.connections: Dict[int, Connection] = {}
         self.is_journal_mode_wal = False
         self.journal_mode_was_changed_to_wal = False
-        self.lock_timeout = lock_timeout
+        super().__init__(
+            pool_size=pool_size,
+            max_overflow=max_overflow,
+            pool_timeout=pool_timeout,
+            max_age=max_age,
+            pre_ping=pre_ping,
+            mutually_exclusive_read_write=self.is_sqlite_memory_mode,
+        )
 
-    def detect_memory_mode(self, db_name: str) -> bool:
+    @staticmethod
+    def detect_memory_mode(db_name: str) -> bool:
         return bool(db_name) and (":memory:" in db_name or "mode=memory" in db_name)
 
-    def transaction(self, commit: bool = False) -> Transaction:
-        c = self.get_connection()
-        return Transaction(c, commit, self.lock)
-
-    def get_connection(self) -> Connection:
-        thread_id = threading.get_ident()
-        try:
-            c = self.connections[thread_id]
-        except KeyError:
-            c = self.create_connection()
-            self.connections[thread_id] = c
-        return c
-
-    def create_connection(self) -> Connection:
+    def _create_connection(self) -> SQLiteConnection:
         # Make a connection to an SQLite database.
         try:
             c = sqlite3.connect(
@@ -146,15 +193,51 @@ class SQLiteDatastore:
         c.row_factory = sqlite3.Row
 
         # Return the connection.
-        return c
+        return SQLiteConnection(sqlite_conn=c, max_age=self.max_age)
 
-    def close_all_connections(self) -> None:
-        for c in self.connections.values():
-            c.close()
-        self.connections.clear()
+
+class SQLiteDatastore:
+    def __init__(
+        self,
+        db_name: str,
+        lock_timeout: Optional[int] = None,
+        pool_size: int = 5,
+        max_overflow: int = 10,
+        pool_timeout: float = 5.0,
+        max_age: Optional[float] = None,
+        pre_ping: bool = False,
+    ):
+        self.pool = SQLiteConnectionPool(
+            db_name=db_name,
+            lock_timeout=lock_timeout,
+            pool_size=pool_size,
+            max_overflow=max_overflow,
+            pool_timeout=pool_timeout,
+            max_age=max_age,
+            pre_ping=pre_ping,
+        )
+
+    @contextmanager
+    def transaction(self, commit: bool) -> Iterator[SQLiteCursor]:
+        with self.get_connection(commit=commit) as conn:
+            with conn.transaction(commit) as curs:
+                yield curs
+
+    @contextmanager
+    def get_connection(self, commit: bool) -> Iterator[SQLiteConnection]:
+        # Using reader-writer interlocking is necessary for in-memory databases,
+        # but also speeds up (and provides "fairness") to file-based databases.
+        conn = self.pool.get_connection(is_writer=commit)
+        try:
+            yield conn
+        finally:
+            self.pool.put_connection(conn)
+
+    def close(self) -> None:
+        self.pool.close()
 
     def __del__(self) -> None:
-        self.close_all_connections()
+        self.close()
 
 
 class SQLiteAggregateRecorder(AggregateRecorder):
@@ -170,7 +253,6 @@ class SQLiteAggregateRecorder(AggregateRecorder):
         self.insert_events_statement = (
             f"INSERT INTO {self.events_table_name} VALUES (?,?,?,?)"
         )
-        # noinspection SqlResolve
         self.select_events_statement = (
             "SELECT * " f"FROM {self.events_table_name} " "WHERE originator_id=? "
         )
@@ -195,16 +277,18 @@ class SQLiteAggregateRecorder(AggregateRecorder):
                 c.execute(statement)
             pass  # for Coverage 5.5 bug with CPython 3.10.0rc1
 
-    def insert_events(self, stored_events: List[StoredEvent], **kwargs: Any) -> None:
+    def insert_events(
+        self, stored_events: List[StoredEvent], **kwargs: Any
+    ) -> Optional[Sequence[int]]:
         with self.datastore.transaction(commit=True) as c:
-            self._insert_events(c, stored_events, **kwargs)
+            return self._insert_events(c, stored_events, **kwargs)
 
     def _insert_events(
         self,
-        c: Cursor,
+        c: BaseCursor,
         stored_events: List[StoredEvent],
         **kwargs: Any,
-    ) -> None:
+    ) -> Optional[Sequence[int]]:
         params = []
         for stored_event in stored_events:
             params.append(
@@ -215,7 +299,8 @@ class SQLiteAggregateRecorder(AggregateRecorder):
                     stored_event.state,
                 )
             )
-        c.executemany(self.insert_events_statement, params)
+        cast(SQLiteCursor, c).executemany(self.insert_events_statement, params)
+        return None
 
     def select_events(
         self,
@@ -242,7 +327,7 @@ class SQLiteAggregateRecorder(AggregateRecorder):
             statement += "LIMIT ? "
             params.append(limit)
         stored_events = []
-        with self.datastore.transaction() as c:
+        with self.datastore.transaction(commit=False) as c:
             c.execute(statement, params)
             for row in c.fetchall():
                 stored_events.append(
@@ -270,10 +355,6 @@ class SQLiteApplicationRecorder(
         self.select_max_notification_id_statement = (
             f"SELECT MAX(rowid) FROM {self.events_table_name}"
         )
-        self.select_notifications_statement = (
-            f"SELECT rowid, * FROM {self.events_table_name} "
-            "WHERE rowid>=? ORDER BY rowid LIMIT ?"
-        )
 
     def construct_create_table_statements(self) -> List[str]:
         statement = (
@@ -288,14 +369,56 @@ class SQLiteApplicationRecorder(
         )
         return [statement]
 
-    def select_notifications(self, start: int, limit: int) -> List[Notification]:
+    def _insert_events(
+        self,
+        c: BaseCursor,
+        stored_events: List[StoredEvent],
+        **kwargs: Any,
+    ) -> Optional[Sequence[int]]:
+        returning = []
+        for stored_event in stored_events:
+            c.execute(
+                self.insert_events_statement,
+                (
+                    stored_event.originator_id.hex,
+                    stored_event.originator_version,
+                    stored_event.topic,
+                    stored_event.state,
+                ),
+            )
+            returning.append(cast(SQLiteCursor, c).lastrowid)
+        return returning
+
+    def select_notifications(
+        self,
+        start: int,
+        limit: int,
+        stop: Optional[int] = None,
+        topics: Sequence[str] = (),
+    ) -> List[Notification]:
         """
         Returns a list of event notifications
         from 'start', limited by 'limit'.
         """
         notifications = []
-        with self.datastore.transaction() as c:
-            c.execute(self.select_notifications_statement, [start, limit])
+
+        params: List[Union[int, str]] = [start]
+        statement = f"SELECT rowid, * FROM {self.events_table_name} " "WHERE rowid>=? "
+
+        if stop is not None:
+            params.append(stop)
+            statement += "AND rowid<=? "
+
+        if topics:
+            params += list(topics)
+            statement += "AND topic IN (%s) " % ",".join("?" * len(topics))
+
+        params.append(limit)
+        statement += "ORDER BY rowid LIMIT ?"
+
+        with self.datastore.transaction(commit=False) as c:
+            c.execute(statement, params)
+
             for row in c.fetchall():
                 notifications.append(
                     Notification(
@@ -313,10 +436,12 @@ class SQLiteApplicationRecorder(
         """
         Returns the maximum notification ID.
         """
-        with self.datastore.transaction() as c:
-            c.execute(self.select_max_notification_id_statement)
-            max_id = c.fetchone()[0] or 0
-        return max_id
+        with self.datastore.transaction(commit=False) as c:
+            return self._max_notification_id(c)
+
+    def _max_notification_id(self, c: BaseCursor) -> int:
+        c.execute(self.select_max_notification_id_statement)
+        return c.fetchone()[0] or 0
 
 
 class SQLiteProcessRecorder(
@@ -329,11 +454,14 @@ class SQLiteProcessRecorder(
         events_table_name: str = "stored_events",
     ):
         super().__init__(datastore, events_table_name)
-        # noinspection SqlResolve
         self.insert_tracking_statement = "INSERT INTO tracking VALUES (?,?)"
         self.select_max_tracking_id_statement = (
             "SELECT MAX(notification_id) FROM tracking WHERE application_name=?"
         )
+        self.count_tracking_id_statement = (
+            "SELECT COUNT(*) FROM tracking WHERE "
+            "application_name=? AND notification_id=?"
+        )
 
     def construct_create_table_statements(self) -> List[str]:
         statements = super().construct_create_table_statements()
@@ -349,18 +477,24 @@ class SQLiteProcessRecorder(
 
     def max_tracking_id(self, application_name: str) -> int:
         params = [application_name]
-        with self.datastore.transaction() as c:
+        with self.datastore.transaction(commit=False) as c:
             c.execute(self.select_max_tracking_id_statement, params)
             max_id = c.fetchone()[0] or 0
         return max_id
 
+    def has_tracking_id(self, application_name: str, notification_id: int) -> bool:
+        params = [application_name, notification_id]
+        with self.datastore.transaction(commit=False) as c:
+            c.execute(self.count_tracking_id_statement, params)
+            return bool(c.fetchone()[0])
+
     def _insert_events(
         self,
-        c: Cursor,
+        c: BaseCursor,
         stored_events: List[StoredEvent],
         **kwargs: Any,
-    ) -> None:
-        super()._insert_events(c, stored_events, **kwargs)
+    ) -> Optional[Sequence[int]]:
+        returning = super()._insert_events(c, stored_events, **kwargs)
         tracking: Optional[Tracking] = kwargs.get("tracking", None)
         if tracking is not None:
             c.execute(
@@ -370,6 +504,7 @@ class SQLiteProcessRecorder(
                     tracking.notification_id,
                 ),
             )
+        return returning
 
 
 class Factory(InfrastructureFactory):
@@ -377,17 +512,19 @@ class Factory(InfrastructureFactory):
     SQLITE_LOCK_TIMEOUT = "SQLITE_LOCK_TIMEOUT"
     CREATE_TABLE = "CREATE_TABLE"
 
-    def __init__(self, application_name: str, env: Mapping[str, str]):
-        super().__init__(application_name, env)
-        db_name = self.getenv(self.SQLITE_DBNAME)
+    def __init__(self, env: Environment):
+        super().__init__(env)
+        db_name = self.env.get(self.SQLITE_DBNAME)
         if not db_name:
             raise EnvironmentError(
                 "SQLite database name not found "
-                "in environment with key "
-                f"'{self.SQLITE_DBNAME}'"
+                "in environment with keys: "
+                f"{', '.join(self.env.create_keys(self.SQLITE_DBNAME))}"
             )
 
-        lock_timeout_str = (self.getenv(self.SQLITE_LOCK_TIMEOUT) or "").strip() or None
+        lock_timeout_str = (
+            self.env.get(self.SQLITE_LOCK_TIMEOUT) or ""
+        ).strip() or None
 
         lock_timeout: Optional[int] = None
         if lock_timeout_str is not None:
@@ -427,4 +564,7 @@ class Factory(InfrastructureFactory):
 
     def env_create_table(self) -> bool:
         default = "yes"
-        return bool(strtobool(self.getenv(self.CREATE_TABLE, default) or default))
+        return bool(strtobool(self.env.get(self.CREATE_TABLE, default) or default))
+
+    def close(self) -> None:
+        self.datastore.close()
